/* memory allocation routines
 * Copyright 1991 Phil Karn, KA9Q
 *
 * Adapted from alloc routine in K&R; memory statistics and interrupt
 * protection added for use with net package. Must be used in place of
 * standard Turbo-C library routines because the latter check for stack/heap
 * collisions. This causes erroneous failures because process stacks are
 * allocated off the heap.
 */

/****************************************************************************
*	$Id: alloc.c 1.5 1997/11/24 22:22:44 Graham Exp $
*	13 Jul 93	1.2		GT	Fix warnings.									*
*	   May 97	1.4		GeK	Fix warning with MULTITASK undef'd, tidied free	*
*	         	   		   	logging											*
*	   Nov 97	1.5		GeK	Added debugging support; added PK's code to		*
*	         	   		   	allocate extra DOS blocks; improved free()		*
*	         	   		   	logging											*
****************************************************************************/

#include <stdio.h>
#include <dos.h>
#include <alloc.h>
#include "global.h"
#include "mbuf.h"
#include "proc.h"
#include "cmdparse.h"
#include "config.h"
#include "socket.h"		/*  GeK - tputs, tputc  */

#undef	MEM_DEBUG		/*  GeK - memory debugging, 'mem debug'  */
#define	DOS_EXTRAMEM	/*  GeK - allocate extra blocks from DOS  */

#ifdef	GEK_DEBUG
void	stktrace (void);	/*  In stktrace.c  */
#endif

static unsigned long Memfail;	/* Count of allocation failures */
static unsigned long Allocs;	/* Total allocations */
static unsigned long Frees;	/* Total frees */
static unsigned long Invalid;	/* Total calls to free with garbage arg */
static unsigned long Intalloc;	/* Calls to malloc with ints disabled */
static unsigned long Intfree;	/* Calls to free with ints disabled */
static int Memwait;		/* Number of tasks waiting for memory */
static unsigned long Yellows;	/* Yellow alert garbage collections */
static unsigned long Reds;	/* Red alert garbage collections */
unsigned long Availmem;		/* Heap memory, ABLKSIZE units */
static unsigned long Morecores;
#ifdef MULTITASK
unsigned Minheap = 16 * 1024;	/* Min free heap when shelling out */
#endif

static unsigned long Sizes[16];

#ifdef	MEM_DEBUG		/*  GeK - memory block debugging  */
static int Memdebug;
static unsigned long Overuse;	/* Total calls to free with overused arg */

/* This debugging pattern MUST be exactly equal in size to the "header"
 * union defined later
 */
static char Debugpat[] = { 0xfe, 0xed, 0xfa, 0xce, 0xde, 0xad, 0xbe, 0xef };

#define MARKER		0x07080900L

static int domdebug (int argc,char *argv[],void *ptr);
#endif	/* MEM_DEBUG */

static int dostat __ARGS((int argc,char *argv[],void *p));
static int dofreelist __ARGS((int argc,char *argv[],void *p));
static int doibufsize __ARGS((int argc,char *argv[],void *p));
static int donibufs __ARGS((int argc,char *argv[],void *p));
static int dothresh __ARGS((int argc,char *argv[],void *p));
static int dosizes __ARGS((int argc,char *argv[],void *p));
#ifdef	MULTITASK		/*  GeK - fix warning  */
static int dominheap __ARGS((int argc,char *argv[],void *p));
#endif

struct cmds Memcmds[] = {
#ifdef	MEM_DEBUG
	{ "debug",		domdebug,	0, 0, NULLCHAR },
#endif
	{ "freelist",	dofreelist,	0, 0, NULLCHAR },
	{ "ibufsize",	doibufsize,	0, 0, NULLCHAR },
#ifdef MULTITASK
	{ "minheap",	dominheap,	0, 0, NULLCHAR },
#endif
	{ "nibufs",	donibufs,	0, 0, NULLCHAR },
	{ "sizes",	dosizes,	0, 0, NULLCHAR },
	{ "status",	dostat,		0, 0, NULLCHAR },
	{ "thresh",	dothresh,	0, 0, NULLCHAR },
	{ NULLCHAR },
};

#ifdef	LARGEDATA
#define	HUGE	huge
#else
#define	HUGE
#endif

struct header {
	struct header HUGE *ptr;
	unsigned long size;
};

typedef struct header HEADER;
#define	NULLHDR	(HEADER HUGE *)NULL

#define	ABLKSIZE	(sizeof (HEADER))

static HEADER HUGE *morecore __ARGS((unsigned nu));

static HEADER Base;
static HEADER HUGE *Allocp = NULLHDR;
static unsigned long Heapsize;

/*  GeK - for checking block size is acceptable  */
#ifdef	MEM_DEBUG
#define MAX_BLOCK_SIZE	((65535U/ABLKSIZE) + 3)		/*  in ABLKSIZE units, 2 is from nu+1 in malloc, 1 for end marker  */
#else
#define MAX_BLOCK_SIZE	((65535U/ABLKSIZE) + 2)		/*  in ABLKSIZE units, 2 is from nu+1 in malloc  */
#endif

#ifdef	DOS_EXTRAMEM
/* Memory blocks obtained from MS-DOS by allocmem() call */
struct sysblock {
	unsigned seg;
	unsigned npar;
};
#define	NSYSBLOCK	5
static struct sysblock Sysblock[NSYSBLOCK];
#endif	/* DOS_EXTRAMEM */

/* Allocate block of 'nb' bytes */
void *
malloc (size_t nb)
{
	HEADER HUGE *p, HUGE *q;
	unsigned nu;
	int i;

	if(!istate())
		Intalloc++;

#ifdef	GEK_DEBUG
	if(nb == 0 || nb > (65535U - (ABLKSIZE-1))) {
		stktrace();
		return NULL;
	}
#endif

	/* Record the size of this request */
	if((i = log2(nb)) >= 0)
		Sizes[i]++;
	
	/* Round up to full block, then add one for header */
#ifdef	MEM_DEBUG
	nu = ((nb + ABLKSIZE - 1) / ABLKSIZE) + 2;
#else
	nu = ((nb + ABLKSIZE - 1) / ABLKSIZE) + 1;
#endif
	if((q = Allocp) == NULLHDR){
		Base.ptr = Allocp = q = &Base;
		Base.size = 1;
	}
	for(p = q->ptr; ; q = p, p = p->ptr){
		if(p->size >= nu){
			/* This chunk is at least as large as we need */

#ifdef	MEM_DEBUG	/*  GeK - check for valid block before using  */
			if (Memdebug) {
				unsigned j, corrupt = 0;

				for (j = 1; j < nu; j++) {
					if (memcmp (p + j, Debugpat, sizeof (Debugpat)) != 0) {
						corrupt = j;
						break;
					}
				}

				if (corrupt) {
					printf ("Corrupt block in malloc: %p %u size=%lu\n", p, corrupt, p->size);
					fflush (stdout);
					stktrace();
					abort();
					if (p != Allocp)
						continue;	/*  for() search loop  */
				}
			}
#endif	/* MEM_DEBUG */

			if(p->size <= nu + 1){
				/* This is either a perfect fit (size == nu)
				 * or the free chunk is just one unit larger.
				 * In either case, alloc the whole thing,
				 * because there's no point in keeping a free
				 * block only large enough to hold the header.
				 */
				q->ptr = p->ptr;
			} else {
				/* Carve out piece from end of entry */
				p->size -= nu;
				p += p->size;
				p->size = nu;
			}
#ifdef	circular
			Allocp = q;
#endif
			p->ptr = p;	/* for auditing */
#ifdef	MEM_DEBUG
			p[p->size-1].ptr = p;			/*  pointer  */
			p[p->size-1].size = MARKER;		/*  signature  */
#endif
			Allocs++;
			Availmem -= p->size;
			p++;
			break;
		}
		if(p == Allocp && ((p = morecore(nu)) == NULLHDR)){
			Memfail++;
			break;
		}
	}
#ifdef	LARGEDATA
	/* On the brain-damaged Intel CPUs in "large data" model,
	 * make sure the pointer's offset field isn't null
	 * (unless the entire pointer is null).
	 * The Turbo C compiler and certain
	 * library functions like strrchr() assume this.
	 */
	if(FP_OFF(p) == 0 && FP_SEG(p) != 0){
		/* Return denormalized but equivalent pointer */
		return (void *)MK_FP(FP_SEG(p)-1,16);
	}
#endif
	return (void *)p;
}
/* Get more memory from the system and put it on the heap */
static HEADER HUGE *
morecore (unsigned nu)
{
	void HUGE *p;
	HEADER HUGE *up;
	unsigned int size;
#ifdef	DOS_EXTRAMEM
	unsigned segp;
	unsigned npar;
	struct sysblock *sp;
	int i;
#endif	/* DOS_EXTRAMEM */

	Morecores++;
	size = nu * ABLKSIZE;
	/* First try to expand our main memory block */
	if ((int)(p = (void HUGE *)sbrk(size)) != -1) {
		up = (HEADER *)p;
		up->size = nu;
		up->ptr = up;	/* satisfy audit */
#ifdef	MEM_DEBUG
		up[nu-1].ptr = up;			/*  satisfy debug overuse audit  */
		up[nu-1].size = MARKER;
#endif
		free(up + 1);
		Heapsize += size;
		Frees--;	/* Nullify increment inside free() */
		return Allocp;
	}
#ifdef	DOS_EXTRAMEM	/*  GeK - from export v951123  */
	/* That failed; the main memory block must have grown into another
	 * allocated block, or something else (e.g., the increase handles
	 * call in ioinit()) must have allocated memory just beyond it.
	 * Allocate or extend an additional memory block.
	 */
	/*  GeK - was size+16, why??  */
	npar = (size+15)/16;	/* Convert size from bytes to paragraphs */
	p = NULL;
	for(sp=Sysblock,i=0;i < NSYSBLOCK;i++,sp++){
		if(sp->npar != 0){
			/* Try to expand this block */
			if(setblock(sp->seg,sp->npar + npar) != -1){
				/* Failed (-1 == SUCCESS; strange!) */
				continue;
			}
			/* Block expansion succeeded */
			p = MK_FP(sp->seg + sp->npar,0);
			sp->npar += npar;
		} else {
			/* Allocate new block */
			if(allocmem(npar,&segp) != -1){
				return NULL;	/* Complete failure */
			}
			/* -1 indicates SUCCESS (strange) */
			sp->seg = segp;
			sp->npar = npar;
			p = MK_FP(segp,0);
		}
		break;
	}
	if(p != (char HUGE *)NULL){
		/* Expand or create succeeded, add to heap */
		up = (HEADER *)p;
		up->size = (npar*16)/ABLKSIZE;
		up->ptr = up;	/* satisfy audit */
#ifdef	MEM_DEBUG
		up[up->size-1].ptr = up;			/*  satisfy debug overuse audit  */
		up[up->size-1].size = MARKER;
#endif
		free(up + 1);
		Heapsize += npar*16;
		Frees--;	/* Nullify increment inside free() */
		return Allocp;
	}
#endif	/* DOS_EXTRAMEM */
	return NULL;
}

/* Put memory block back on heap */
void
free (void *blk)
{
	HEADER HUGE *p, HUGE *q;
	unsigned long *ptr;
	static char	freewarn[] = "free: %s %p (%08lx size=%08lx) caller %04x:%04x proc %s%c";
	static char	freewarn1[] = "WARNING! invalid pointer";
#ifdef	MEM_DEBUG
	static char freewarn2[] = "overused buffer";
#endif
	
	if(!istate())
		Intfree++;
	if(blk == NULL)
		return;		/* Required by ANSI */
	Frees++;
	p = ((HEADER HUGE *)blk) - 1;
	/* Audit check */	/*  GeK - added check for expected block size range  */
	if(p->ptr != p || p->size < 2 || p->size > MAX_BLOCK_SIZE){
		ptr = ((unsigned long *)&blk) - 1;
		printf (freewarn, freewarn1, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\n');
		fflush (stdout);
#ifdef	GEK_DEBUG
		stktrace();
#endif
		Invalid++;
		log (-1, freewarn, freewarn1, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\0');
		abort();
		return;
	}
#ifdef	MEM_DEBUG
	if (p[p->size-1].ptr != p || p[p->size-1].size != MARKER) {
		ptr = ((unsigned long *)&blk) - 1;
		printf (freewarn, freewarn2, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\n');
		fflush (stdout);
#ifdef	GEK_DEBUG
		stktrace();
#endif
		Overuse++;
		log (-1, freewarn, freewarn2, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\0');
		abort();
		return;		
	}
#endif
	Availmem += p->size;
#ifdef	MEM_DEBUG
	if(Memdebug){
		unsigned i;

		/* Fill data area with pattern to detect later overwrites */
		for(i=1;i<p->size;i++)
			memcpy (p+i, Debugpat, sizeof (Debugpat));
	}
#endif	/* MEM_DEBUG */
	/* Search the free list looking for the right place to insert */
	for(q = Allocp; !(p > q && p < q->ptr); q = q->ptr){
		/* Highest address on circular list? */
		if(q >= q->ptr && (p > q || p < q->ptr))
			break;
	}
	if(p + p->size == q->ptr){
		/* Combine with front of this entry */
		p->size += q->ptr->size;
		p->ptr = q->ptr->ptr;
#ifdef	MEM_DEBUG
		if (Memdebug)
			memcpy (q->ptr, Debugpat, sizeof (Debugpat));
#endif
	} else {
		/* Link to front of this entry */
		p->ptr = q->ptr;
	}
	if(q + q->size == p){
		/* Combine with end of this entry */
		q->size += p->size;
		q->ptr = p->ptr;
#ifdef	MEM_DEBUG
		if(Memdebug){
			memcpy (p, Debugpat, sizeof (Debugpat));
		}
#endif
	} else {
		/* Link to end of this entry */
		q->ptr = p;
	}
#ifdef	circular
	Allocp = q;
#endif
	if((volatile)Memwait != 0)
		psignal(&Memwait,0);
}

#ifdef	notdef	/* Not presently used */
/* Move existing block to new area */
void *
realloc(area,size)
void *area;
unsigned size;
{
	unsigned osize;
	HEADER HUGE *hp;
	char HUGE *cp;

	hp = ((HEADER *)area) - 1;
	osize = (hp->s.size -1) * ABLKSIZE;

	free(area);
	if((cp = malloc(size)) != NULL && cp != area)
		memcpy((char *)cp,(char *)area,size>osize? osize : size);
	return cp;
}
#endif

/* Allocate block of cleared memory */
void *
calloc (size_t nelem,	/* Number of elements */
        size_t size)	/* Size of each element */
{
	unsigned i;
	char *cp;

	i = nelem * size;
	if((cp = malloc(i)) != NULL)
		memset(cp,0,i);
	return cp;
}

/* Version of malloc() that waits if necessary for memory to become available */
void *
mallocw (size_t nb)
{
	void *p;

	while((p = malloc(nb)) == NULL){
		Memwait++;
		pwait(&Memwait);
		Memwait--;
	}
	return p;
}

/* Version of calloc that waits if necessary for memory to become available */
void *
callocw (size_t nelem,	/* Number of elements */
         size_t size)	/* Size of each element */
{
	size_t i;
	char *cp;

	i = nelem * size;
	cp = mallocw(i);
	memset(cp,0,i);
	return cp;
}

/* Return available memory on our heap plus available system memory */
unsigned long
availmem (void)
{
	return Availmem * ABLKSIZE + coreleft();
}

/* Print heap stats */
static int
dostat (int argc, char *argv[], void *envp)
{
#ifdef	DOS_EXTRAMEM
	struct sysblock *sp;
	int i;
#endif	/* DOS_EXTRAMEM */

	tprintf("heap size %lu avail %lu (%lu%%) morecores %lu coreleft %lu\n",
	 Heapsize,Availmem * ABLKSIZE,100L*Availmem*ABLKSIZE/Heapsize,
	 Morecores,coreleft());
#ifdef	DOS_EXTRAMEM
 	if(Sysblock[0].npar != 0){
		tputs ("Extra blocks:");
		for(i=0,sp=Sysblock;i< NSYSBLOCK;i++,sp++){
			if(sp->npar == 0)
				break;
			tprintf (" (%04x0-%04xf)",sp->seg,sp->seg+sp->npar-1);
		}
		tputc ('\n');
	}
#endif	/* DOS_EXTRAMEM */
#ifdef	MEM_DEBUG
	tprintf("allocs %lu frees %lu (diff %lu) alloc fails %lu invalid frees %lu overused %lu\n",
		Allocs,Frees,Allocs-Frees,Memfail,Invalid,Overuse);
#else
	tprintf("allocs %lu frees %lu (diff %lu) alloc fails %lu invalid frees %lu\n",
		Allocs,Frees,Allocs-Frees,Memfail,Invalid);
#endif
	tprintf("interrupts-off calls to malloc %lu free %lu\n",Intalloc,Intfree);
	tprintf("garbage collections yellow %lu red %lu\n",Yellows,Reds);
	iqstat();
	return 0;
}

/* Print heap free list */
static int
dofreelist (int argc, char *argv[], void *envp)
{
	HEADER HUGE *p;
	int i = 0;
#ifdef	MEM_DEBUG
	unsigned j;
	unsigned corrupt;
#endif

	for(p = Base.ptr; p != (HEADER HUGE *)&Base; p = p->ptr){
#ifdef	MEM_DEBUG
		corrupt = 0;
		if (Memdebug) {
			for (j=1; j<p->size; j++) {
				if (memcmp (p+j, Debugpat, sizeof (Debugpat)) != 0) {
					corrupt = j;
					break;
				}
			}
		}
#endif	/* MEM_DEBUG */
		tprintf ("%p %6lu", p, p->size * ABLKSIZE);		
#ifdef	MEM_DEBUG
		if (corrupt) {
			tprintf (" C:%u", corrupt);
			i = 3;
		}
#endif	/* MEM_DEBUG */
		if(++i == 4){
			i = 0;
			if(tputc('\n') == EOF)
				return 0;
		} else
			tputs(" | ");
		/*  GeK - flush screen o/p to try and avoid disturbing list  */
		usflush (Curproc->output);
	}
	if(i != 0)
		tputc('\n');

	return 0;
}

static int
dosizes (int argc, char *argv[], void *p)
{
	int i;

	for(i=0;i<16;i += 4){
		tprintf("N>=%5u:%7ld| N>=%5u:%7ld| N>=%5u:%7ld| N>=%5u:%7ld\n",
		 1<<i,Sizes[i],	2<<i,Sizes[i+1],
		 4<<i,Sizes[i+2],8<<i,Sizes[i+3]);
	}
	return 0;
}

int
domem (int argc, char *argv[], void *p)
{
	return subcmd(Memcmds,argc,argv,p);
}

#ifdef	MEM_DEBUG
static int
domdebug (int argc, char *argv[], void *ptr)
{
	int prev,i_state;
	unsigned j;

	HEADER HUGE *p;

	prev = Memdebug;
	setbool(&Memdebug,"Heap debugging",argc,argv);
	if(prev == 1 || Memdebug == 0)
		return 0;

	/* Turning debugging on; reinitialize free areas to debug pattern */
	i_state = dirps();
	for (p = Base.ptr; p != (HEADER HUGE *)&Base; p = p->ptr) {
		for (j = 1; j < p->size; j++) {
			memcpy (p + j, Debugpat, sizeof (Debugpat));
		}
	}
	restore (i_state);
	return 0;
}
#endif	/* MEM_DEBUG */

static int
donibufs (int argc, char *argv[], void *p)
{
	return setint(&Nibufs,"Interrupt pool buffers",argc,argv);
}

static int
doibufsize (argc,argv,p)
int argc;
char *argv[];
void *p;
{
	return setuns(&Ibufsize,"Interrupt buffer size",argc,argv);
}

static int
dothresh (int argc, char *argv[], void *p)
{
	return setlong(&Memthresh,"Free memory threshold (bytes)",argc,argv);
}

/* Background memory compactor, used when memory runs low */
void
gcollect (int i, void *v1, void *v2)	/* Args not used */
{
	void (**fp)(int colour);
	int red;

	for(;;){
		pause(1000L);	/* Run every second */
		/* If memory is low, collect some garbage. If memory is VERY
		 * low, invoke the garbage collection routines in "red" mode.
		 */
		if(availmem() < Memthresh){
			if(availmem() < Memthresh/2){
				red = 1;
				Reds++;
			} else {
				red = 0;
				Yellows++;
			}
			for(fp = Gcollect;*fp != NULL;fp++)
				(**fp)(red);
		}
	}
}

#ifdef MULTITASK
static int
dominheap (int argc, char *argv[], void *p)
{
	return setuns(&Minheap,"Minimum free heap when shelled out",argc,argv);
}
#endif

#ifdef	GEK_DEBUG
int
validate_pointer (void *blk, const char * const desc)
{
	HEADER HUGE *p;
	unsigned long *ptr;
	static char	freewarn[] = "vp: %s %s %p (%08lx size=%08lx) caller %04x:%04x proc %s%c";
	static char	freewarn1[] = "invalid pointer";
#ifdef	MEM_DEBUG
	static char freewarn2[] = "overused buffer";
#endif

	if (blk == NULL)
	{
		static char	vp1[] = "vp: NULL pointer: %s%c";
		printf (vp1, desc, '\n');
		fflush (stdout);
		log (-1, vp1, desc, '\0');
		return 0;
	}

	p = ((HEADER HUGE *)blk) - 1;
	/* Audit check */	/*  GeK - added check for expected block size range  */
	if(p->ptr != p || p->size < 2 || p->size > MAX_BLOCK_SIZE) {
		ptr = ((unsigned long *)&blk) - 1;
		printf (freewarn, freewarn1, desc, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\n');
		fflush (stdout);
		log (-1, freewarn, freewarn1, desc, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\0');
		return 0;
	}
#ifdef	MEM_DEBUG
	if (p[p->size-1].ptr != p || p[p->size-1].size != MARKER) {
		ptr = ((unsigned long *)&blk) - 1;
		printf (freewarn, freewarn2, desc, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\n');
		fflush (stdout);
		log (-1, freewarn, freewarn2, desc, blk, p->ptr, p->size, FP_SEG(*ptr)-(_psp+0x0010), FP_OFF(*ptr), Curproc->name, '\0');
		return 0;		
	}
#endif
	return 1;
}
#endif	/* GEK_DEBUG */
