/*
 *	httpd.c - RFC2068 HTTP daemon for KA9Q
 *
 *	Graham E. Kinns	<Graham@grempc.demon.co.uk>	Oct '97
 *
 *	$Id: $
 */

#include <ctype.h>		/*  isdigit(), isspace()  */
#include "global.h"
#include "proc.h"
#include "socket.h"
#include "domain.h"		/*  resolve_a()  */
#include "netuser.h"	/*  inet_ntoa()  */

#ifndef	IPPORT_HTTP
#  define IPPORT_HTTP 80
#endif

/*  What's sent in the Server: header  */
#define SERVER	"GremNOS-httpd/1.0"

#define HTTPD_DEFAULT_IDLE	300	/*  5 minute idle out on receive  */
#define HTTPD_BUF_LEN		512

/*  Externals declated in smtpserv.c  */
extern char *Days[7];
extern char *Months[12];

/*  global function prototypes in httpd.c  */
int	httpd_start (int argc, char *argv[], void *p);
int	httpd_stop  (int argc, char *argv[], void *p);

/*  prototypes for local functions  */
static void	httpd (int s, void *unused, void *p);
static char *http_date (time_t *t);

/*	local global variable  */
static int	Shttpd = -1;	/*  Prototype socket for service  */
static char *redirect_url;	/*  Redirect Location: URL issued by httpd  */

/*  Start HTTP daemon  */
int
httpd_start (int argc, char *argv[], void *p)
{
	struct sockaddr_in	lsocket;
	int					s;

	if (Shttpd != -1)
		return (0);		/*  httpd already running  */

	psignal (Curproc, 0);	/* Don't keep the parser waiting */
	chname (Curproc, "HTTP listener");

	if (argc > 1)
	{
		redirect_url = strdup (argv[1]);
	}
	else
	{
		redirect_url = mallocw (12 + strlen (Hostname) + 1);
		sprintf (redirect_url, "http://www.%s/", Hostname);
	}

	lsocket.sin_family = AF_INET;
	lsocket.sin_addr.s_addr = INADDR_ANY;
	lsocket.sin_port = IPPORT_HTTP;

	Shttpd = socket (AF_INET, SOCK_STREAM, 0);
	if (Shttpd == -1)
	{
		log (-1, "httpd_start: %s() failed, errno=%d", "socket", errno);
		return (0);		/* socket failed */
	}
	
	if (bind (Shttpd, (char *) &lsocket, sizeof (lsocket)) == -1)
	{
		log (-1, "httpd_start: %s() failed, errno=%d", "bind", errno);
		httpd_stop (argc, argv, p);
		return (0);		/* bind failed */
	}

	if (listen (Shttpd, 1) == -1)		/*  Accept multiple connections  */
	{
		log (-1, "httpd_start: %s() failed, errno=%d", "listen", errno);
		httpd_stop (argc, argv, p);
		return (0);		/* listen failed */
	}

	for (;;)
	{
		s = accept (Shttpd, NULL, (int *) NULL);
		if (s == -1)
			break;	/* Service is shutting down */

		/* Spawn a server */
		newproc ("httpd", 512, httpd, s, NULL, NULL, 0);
	}

	return (0);
}

int
httpd_stop (int argc, char *argv[], void *p)
{
	if (Shttpd != -1)
	{
		close_s (Shttpd);
		Shttpd = -1;
		free (redirect_url);		/*  Ick - could still be in use below  */
	}

	return (0);
}

static void
httpd (int s, void *unused, void *p)
{
	char	*buf;
	int		seen_first_line = FALSE;
	struct sockaddr_in	client_ip;   /*  GeK - used in HELO  */
#define client_addr client_ip.sin_addr.s_addr
	int		socksize = SOCKSIZE;
	char	*client_hostname;

	buf = mallocw (HTTPD_BUF_LEN);
	sockmode (s, SOCK_ASCII);
	sockowner (s, Curproc);

	client_addr = 0;
	getpeername (s, (char *) &client_ip, &socksize);
	client_hostname = resolve_a (client_addr, FALSE);
	if (client_hostname != NULL)
	{
		int	len;

		/*  Trim trailing dot  */
		len = strlen (client_hostname);
		if (len > 1 && client_hostname[len-1] == '.')
			client_hostname[len-1] = '\0';
	}
	else
	{
		client_hostname = mallocw (18);
		sprintf (client_hostname, "[%s]", inet_ntoa (client_addr));
	}

	log (s, "open httpd from %s", client_hostname);

	for (;;)
	{
		/*  Start timer  */
		alarm ((int32) HTTPD_DEFAULT_IDLE * 1000L);
		if (recvline (s, buf, HTTPD_BUF_LEN) == -1)
		{
			char	*cp;

			alarm (0L);
			cp = sockerr (s);
			log (s, "httpd recv failed: %s", cp != NULL ? cp : "timeout");
			goto quit;
		}
		alarm (0L);

		rip (buf);
		if (buf[0] == '\0')
			break;		/*  End of sent header  */

		if (!seen_first_line)
		{
			/*  Show GET, HEAD, etc. request  */
			tprintf ("httpd: %s: %s\n", client_hostname, buf);
			seen_first_line = TRUE;
		}

		log (s, "httpd: '%s'", buf);		/*  log who's asking what  */
	}

	/*  Send response  */
	usputs   (s, "HTTP/1.0 302 Object Moved\n");
	usputs   (s, "Server: " SERVER "\n");
	usprintf (s, "Date: %s\n", http_date (NULL));
	usprintf (s, "Location: %s\n", redirect_url);
	usputs   (s, "\n");

quit:
	log (s, "close httpd");
	close_s (s);
	free (buf);
	free (client_hostname);
}

/*  Returns ptr to RFC compliant timestamp, GMT based  */
static char *
http_date (time_t *t)
{
	static char	buf[32];
	time_t	time_now;
	struct tm	*gmt;
	
	time_now = (t == NULL) ? time (NULL) : *t;
	gmt = gmtime (&time_now);
	
	sprintf (buf, "%s, %02u %s %04u %02u:%02u:%02u GMT",
	         Days[gmt->tm_wday],
	         gmt->tm_mday,
	         Months[gmt->tm_mon],
	         gmt->tm_year + 1900,
	         gmt->tm_hour,
	         gmt->tm_min,
	         gmt->tm_sec);

	return (buf);
}
