
/*
 * @(#)wildmat.c 1.3 87/11/06	Public Domain.
 *
From: rs@mirror.TMC.COM (Rich Salz)
Newsgroups: net.sources
Subject: Small shell-style pattern matcher
Message-ID: <596@mirror.TMC.COM>
Date: 27 Nov 86 00:06:40 GMT

There have been several regular-expression subroutines and one or two
filename-globbing routines in mod.sources.  They handle lots of
complicated patterns.  This small piece of code handles the *?[]\
wildcard characters the way the standard Unix(tm) shells do, with the
addition that "[^.....]" is an inverse character class -- it matches
any character not in the range ".....".  Read the comments for more
info.

For my application, I had first ripped off a copy of the "glob" routine
from within the find(1) source, but that code is bad news:  it recurses
on every character in the pattern.  I'm putting this replacement in the
public domain.  It's small, tight, and iterative.  Compile with -DTEST
to get a test driver.  After you're convinced it works, install in
whatever way is appropriate for you.

I would like to hear of bugs, but am not interested in additions; if I
were, I'd use the code I mentioned above.
*/
/*
**  Do shell-style pattern matching for ?, \, [], and * characters.
**  Might not be robust in face of malformed patterns; e.g., "foo[a-"
**  could cause a segmentation violation.
**
**  Written by Rich $alz, mirror!rs, Wed Nov 26 19:03:17 EST 1986.
*/

/*
 * Modified 6Nov87 by John Gilmore (hoptoad!gnu) to return a "match"
 * if the pattern is immediately followed by a "/", as well as \0.
 * This matches what "tar" does for matching whole subdirectories.
 *
 * The "*" code could be sped up by only recursing one level instead
 * of two for each trial pattern, perhaps, and not recursing at all
 * if a literal match of the next 2 chars would fail.
 */

/* Modified by Anders Klemets to take an array of pointers as an optional
   argument. Each part of the string that matches '*' is returned as a
   null-terminated, malloced string in this array.
 */

/*  GeK - from wildmat(3) man page

     Robert  Elz  <kre@munnari.oz.au>  added  minus  sign  and  close  bracket
     handling in June, 1991.
 */

/*  GeK - abort code comment from INN 1.4unoff4 and INN 1.5.1 lib/wildmat.c

**  Special thanks to Lars Mathiesen <thorinn@diku.dk> for the ABORT code.
**  This can greatly speed up failing wildcard patterns.  For example:
**      pattern: -*-*-*-*-*-*-12-*-*-*-m-*-*-*
**      text 1:  -adobe-courier-bold-o-normal--12-120-75-75-m-70-iso8859-1
**      text 2:  -adobe-courier-bold-o-normal--12-120-75-75-X-70-iso8859-1
**  Text 1 matches with 51 calls, while text 2 fails with 54 calls.  Without
**  the ABORT code, it takes 22310 calls to fail.  Ugh.  The following
**  explanation is from Lars:
**  The precondition that must be fulfilled is that DoMatch will consume
**  at least one character in text.  This is true if *p is neither '*' nor
**  '\0'.)  The last return has ABORT instead of FALSE to avoid quadratic
**  behaviour in cases like pattern "*a*b*c*d" with text "abcxxxxx".  With
**  FALSE, each star-loop has to run to the end of the text; with ABORT
**  only the last one does.
**
**  Once the control of one instance of DoMatch enters the star-loop, that
**  instance will return either TRUE or ABORT, and any calling instance
**  will therefore return immediately after (without calling recursively
**  again).  In effect, only one star-loop is ever active.  It would be
**  possible to modify the code to maintain this context explicitly,
**  eliminating all recursive calls at the cost of some complication and
**  loss of clarity (and the ABORT stuff seems to be unclear enough by
**  itself).  I think it would be unwise to try to get this into a
**  released version unless you have a good test data base to try it out
**  on.
*/

/****************************************************************************
*	$Id: wildmat.c 1.4 1997/06/25 17:39:08 Graham Exp $
*	14 Jul 93	1.2		GT	Fix warnings.									*
*	   Nov 96	1.3		GeK	Additions from wildmat(3) (INN1.4unoff4) to		*
*	         	   		   	handle '-' & ']' within [...]					*
*	   Jun 97	1.4		GeK	Added ABORT code from INN wildmat, optimized	*
*	         	   		   	to remove need for (broken) Star() function,	*
*	         	   		   	core function renamed DoMatch(), added new 		*
*	         	   		   	wildmat() user entry point, configurable		*
*	         	   		   	conditionals, tidied test routine				*
****************************************************************************/

#include "global.h"

#ifdef	TEST
#include <stdio.h>
#include <stdlib.h>
#define mallocw	malloc
#endif	/* TEST */

#undef	OPTIMIZE_JUST_STAR		/*  Optimize '*', IMHO not needed with KA9Q  */
#undef	MATCH_TAR_PATTERN		/*  Match trailing / for tar  */
#undef	REDUCE_CONSECUTIVE_STARS	/*  GeK - reduce ** to *  */

#define	ABORT	(-1)

#ifdef	notdef	/*  GeK - removed Star fn  */
static int Star __ARGS((char *s,char *p,char **argv));
#endif

static int DoMatch (const char *s, const char *p, char **argv);

#ifdef	notdef	/*  GeK - removed Star fn  */
static int
Star(s,p,argv)
register char *s;
register char *p;
register char **argv;
{
	char *cp = s;
	while (wildmat(cp, p, argv) == FALSE)
		/*
		 *  GeK - this is a source of a string overrun if
		 *  Star() is called with *s == '\0' and the first
		 *  wildmat() returns FALSE then cp falls off the
		 *  end of s.
		 */
		if(*++cp == '\0')
			return -1;
	return ((int) (cp - s));
}
#endif	/* notdef */

static int
DoMatch (const char *s, const char *p, char **argv)
{
	register int last;
	register int matched;
	register int reverse;
	register int cnt;

#ifdef	TEST
	printf ("  DoMatch p='%s' s='%s'\n", p, s);
#endif

	for(; *p; s++,p++){
		if (*s == '\0' && *p != '*')
			return ABORT;
		switch(*p){
		case '\\':
			/* Literal match with following character; fall through. */
			p++;
		default:
			if(*s != *p)
				return FALSE;
			continue;
		case '?':
			/* Match anything. */
			if(*s == '\0')
				return FALSE;
			continue;
		case '*':
			if (argv == NULL) {
#ifdef	REDUCE_CONSECUTIVE_STARS		/*  GeK  */
				while (*++p == '*')
					/* Consecutive stars act just like one. */
					continue;
#else	/* !REDUCE_CONSECUTIVE_STARS */
				p++;
#endif
				if (*p == '\0')
					/* Trailing star matches everything. */
					return TRUE;

				/*
				 *  GeK - reinstated code from INN's wildmat(3)
				 *  same functionality as Star() without the overhead
				 */
				while (*s)
					if ((matched = DoMatch (s++, p, NULL)) != FALSE)
						return matched;
				return ABORT;
			}

#ifdef	REDUCE_CONSECUTIVE_STARS		/*  GeK  */
			while (*++p == '*') {
				/* Consecutive stars act just like one. */
				/*  GeK - store empty string  */
				*argv = mallocw (1);
				**(argv++) = '\0';
			}
#else	/* !REDUCE_CONSECUTIVE_STARS */
			p++;
#endif
			if (*p == '\0') {
				/* Trailing star matches everything. */
				cnt = strlen (s);
			} else {
				const char *cp = s;

				/*
				 *  GeK - rewritten Star() code, same functionality as
				 *  Star() and above, rearranged to fall through on TRUE.
				 *  Returns ABORT on failure
				 */
				matched = ABORT;
				while (*cp && (matched = DoMatch (cp, p, argv+1)) == FALSE)
					cp++;	/*  Increment inside loop so cnt is correct below  */
				if (matched != TRUE)
					return ABORT;
				cnt = (int) (cp - s);
			}
			*argv = mallocw(cnt+1);
			strncpy(*argv,s,cnt);
			*(*argv + cnt) = '\0';
			return TRUE;
		case '[':
			/* [^....] means inverse character class. */
			reverse = (p[1] == '^') ? TRUE : FALSE;
			if(reverse)
				p++;
			matched = FALSE;
			if (p[1] == ']' || p[1] == '-')
				if (*++p == *s)
					matched = TRUE;
			for (last = *p; *++p && *p != ']'; last = *p) {
				/* This next line requires a good C compiler. */
				if ((*p == '-' && p[1] != ']') ? *s <= *++p && *s >= last : *s == *p)
					matched = TRUE;
			}
			if(matched == reverse)
				return FALSE;
			continue;
		}
	}
#ifdef	MATCH_TAR_PATTERN
	/* For "tar" use, matches that end at a slash also work. --hoptoad!gnu */
	if (*s == '/')
		return TRUE;
#endif	/* MATCH_TAR_PATTERN */
	return (*s == '\0');
}

int
wildmat (char *text, char *p, char **argv)
{
#ifdef	OPTIMIZE_JUST_STAR
	if (p[0] == '*' && p[1] == '\0') {
		if (argv != NULL)
			*argv = strdup (text);
		return TRUE;
	}
#endif	/* OPTIMIZE_JUST_STAR */

	return (DoMatch (text, p, argv) == TRUE);
}


#ifdef	TEST
int
main (void)
{
	char pattern[80];
	char text[80];
	char *argv[80];
	int cnt;
    
	while (TRUE){
		printf("Enter pattern:  ");
		fflush(stdout);
		if(gets(pattern) == NULL || *pattern == '\0')
			break;
		while (TRUE){
			memset (argv, 0, 80*sizeof(char *));
			printf("Enter text:  ");
			fflush(stdout);
			if(gets(text) == NULL)
				exit(0);
			if(text[0] == '\0')
				/* Blank line; go back and get a new pattern. */
				break;
			printf("      %d\n", wildmat(text, pattern, argv));
			for(cnt = 0; argv[cnt] != NULLCHAR; ++cnt){
				printf("String %d is: '%s'\n",cnt,argv[cnt]);
				free(argv[cnt]);
			}
		}
	}
	return (0);
}
#endif	/* TEST */
